@@ -4,9 +4,9 @@
 #
 # RFC 1700, 791, 793, 768, 792... - IP, TCP, UDP and ICMP...
 #
-# (c)2008-2010 Matthias Eckert, Thomas Hungenberg
+# (c)2008-2018 Matthias Eckert, Thomas Hungenberg
 #
-# Version 0.20   (2010-04-10)
+# Version 0.21   (2018-06-05)
 #
 # For history/changelog see bottom of this file.
 #
@@ -17,7 +17,8 @@ package INetSim::Redirect;
 use strict;
 use warnings;
 
-use IPTables::IPv4::IPQueue qw(:constants);
+use nfqueue;
+use Socket qw(AF_INET);
 
 
 my $serviceName = "redirect";
@@ -54,7 +55,7 @@ my $IN_DEV;
 my $OUT_DEV;
 
 my $PID = $$;
-my $ipq;
+my $nfq;
 
 my %type = (	0				=> "echo-reply",
 		3				=> "destination-unreachable",
@@ -342,9 +343,9 @@ sub create_chains {
     # create chain for changing the source ip/port of packets
     &ipt("-t nat -N INetSim_SNAT_$PID");
     # add rule to redirect all packets with state NEW to userspace
-    &ipt("-t mangle -A INetSim_$PID -m state --state NEW -j QUEUE");
+    &ipt("-t mangle -A INetSim_$PID -m state --state NEW -j NFQUEUE");
     # add rule to redirect all icmp timestamp replies to userspace
-    &ipt("-t mangle -A INetSim_$PID -p icmp --icmp-type 14 -j QUEUE");
+    &ipt("-t mangle -A INetSim_$PID -p icmp --icmp-type 14 -j NFQUEUE");
     # now redirect all packets to inetsim chains
     # queue
     &ipt("-t mangle -A PREROUTING -j INetSim_$PID");
@@ -402,8 +403,8 @@ sub delete_chains {
     &ipt("-t nat -F INetSim_REDIRECT_$PID");
     &ipt("-t nat -X INetSim_REDIRECT_$PID");
     # delete chains in mangle table
-    &ipt("-t mangle -D INetSim_$PID -p icmp --icmp-type 14 -j QUEUE");
-    &ipt("-t mangle -D INetSim_$PID -m state --state NEW -j QUEUE");
+    &ipt("-t mangle -D INetSim_$PID -p icmp --icmp-type 14 -j NFQUEUE");
+    &ipt("-t mangle -D INetSim_$PID -m state --state NEW -j NFQUEUE");
     &ipt("-t mangle -F INetSim_$PID");
     &ipt("-t mangle -X INetSim_$PID");
     # set original value for nf_conntrack
@@ -481,10 +482,10 @@ sub process_packet_icmp {
             if ($changeTTL) {
                 $ttl_set = int(rand(30) + 34);
                 &ipt("-t mangle -I INetSim_$PID -s $SRC_IP -d $DST_IP -p $PROTO --icmp-type $TYPE -m connmark --mark 0 -j CONNMARK --set-mark $ttl_set");
-                &INetSim::Log::SubLog("[$SRC_IP:$type{$TYPE}] Translating $PROTO connections from host '$SRC_IP' ($MAC), source changed from '$SRC_IP' to '$externalAddress', destination changed from '$DST_IP:$type{$TYPE}' to '$real_ip:$type{$TYPE}', TTL set to $ttl_set.", $serviceName, $$);
+                &INetSim::Log::SubLog("[$SRC_IP:$type{$TYPE}] Translating $PROTO connections from host '$SRC_IP'" . (($MAC) ? " ($MAC)" : "") . ", source changed from '$SRC_IP' to '$externalAddress', destination changed from '$DST_IP:$type{$TYPE}' to '$real_ip:$type{$TYPE}', TTL set to $ttl_set.", $serviceName, $$);
             }
             else {
-                &INetSim::Log::SubLog("[$SRC_IP:$type{$TYPE}] Translating $PROTO connections from host '$SRC_IP' ($MAC), source changed from '$SRC_IP' to '$externalAddress', destination changed from '$DST_IP:$type{$TYPE}' to '$real_ip:$type{$TYPE}'.", $serviceName, $$);
+                &INetSim::Log::SubLog("[$SRC_IP:$type{$TYPE}] Translating $PROTO connections from host '$SRC_IP'" . (($MAC) ? " ($MAC)" : "") . ", source changed from '$SRC_IP' to '$externalAddress', destination changed from '$DST_IP:$type{$TYPE}' to '$real_ip:$type{$TYPE}'.", $serviceName, $$);
             }
             return 1;
         }
@@ -524,10 +525,10 @@ sub process_packet_icmp {
         if ($changeTTL) {
             $ttl_set = int(rand(30) + 34);
             &ipt("-t mangle -I INetSim_$PID -s $SRC_IP -d $DST_IP -p $PROTO --icmp-type $TYPE -m connmark --mark 0 -j CONNMARK --set-mark $ttl_set");
-            &INetSim::Log::SubLog("[$SRC_IP:$type{$TYPE}] Forwarding $PROTO connections from host '$SRC_IP' ($MAC) to destination '$DST_IP:$type{$TYPE}', TTL set to $ttl_set.", $serviceName, $$);
+            &INetSim::Log::SubLog("[$SRC_IP:$type{$TYPE}] Forwarding $PROTO connections from host '$SRC_IP'" . (($MAC) ? " ($MAC)" : "") . " to destination '$DST_IP:$type{$TYPE}', TTL set to $ttl_set.", $serviceName, $$);
         }
         else {
-            &INetSim::Log::SubLog("[$SRC_IP:$type{$TYPE}] Forwarding $PROTO connections from host '$SRC_IP' ($MAC) to destination '$DST_IP:$type{$TYPE}'.", $serviceName, $$);
+            &INetSim::Log::SubLog("[$SRC_IP:$type{$TYPE}] Forwarding $PROTO connections from host '$SRC_IP'" . (($MAC) ? " ($MAC)" : "") . " to destination '$DST_IP:$type{$TYPE}'.", $serviceName, $$);
         }
         return 1;
     }
@@ -540,7 +541,7 @@ sub process_packet_icmp {
         else {
             &ipt("-t mangle -I INetSim_$PID -s $SRC_IP -d $DST_IP -p $PROTO --icmp-type $TYPE -j MARK --set-mark 1");
         }
-        &INetSim::Log::SubLog("[$SRC_IP:$type{$TYPE}] No rule for $PROTO connections from host '$SRC_IP' ($MAC) to destination '$DST_IP:$type{$TYPE}' - ignored.", $serviceName, $$);
+        &INetSim::Log::SubLog("[$SRC_IP:$type{$TYPE}] No rule for $PROTO connections from host '$SRC_IP'" . (($MAC) ? " ($MAC)" : "") . " to destination '$DST_IP:$type{$TYPE}' - ignored.", $serviceName, $$);
     }
     return 0;
 }
@@ -606,10 +607,10 @@ sub process_packet_tcpudp {
             if ($changeTTL) {
                 $ttl_set = int(rand(30) + 34);
                 &ipt("-t mangle -I INetSim_$PID -s $SRC_IP -d $DST_IP -p $PROTO --dport $DST_PORT -m connmark --mark 0 -j CONNMARK --set-mark $ttl_set");
-                &INetSim::Log::SubLog("[$SRC_IP:$SRC_PORT] Translating $PROTO connections from host '$SRC_IP' ($MAC), source changed from '$SRC_IP' to '$externalAddress', destination changed from '$DST_IP:$DST_PORT' to '$real_ip:$real_port', TTL set to $ttl_set.", $serviceName, $$);
+                &INetSim::Log::SubLog("[$SRC_IP:$SRC_PORT] Translating $PROTO connections from host '$SRC_IP'" . (($MAC) ? " ($MAC)" : "") . ", source changed from '$SRC_IP' to '$externalAddress', destination changed from '$DST_IP:$DST_PORT' to '$real_ip:$real_port', TTL set to $ttl_set.", $serviceName, $$);
             }
             else {
-                &INetSim::Log::SubLog("[$SRC_IP:$SRC_PORT] Translating $PROTO connections from host '$SRC_IP' ($MAC), source changed from '$SRC_IP' to '$externalAddress', destination changed from '$DST_IP:$DST_PORT' to '$real_ip:$real_port'.", $serviceName, $$);
+                &INetSim::Log::SubLog("[$SRC_IP:$SRC_PORT] Translating $PROTO connections from host '$SRC_IP'" . (($MAC) ? " ($MAC)" : "") . ", source changed from '$SRC_IP' to '$externalAddress', destination changed from '$DST_IP:$DST_PORT' to '$real_ip:$real_port'.", $serviceName, $$);
             }
             return 1;
         }
@@ -632,10 +633,10 @@ sub process_packet_tcpudp {
         if ($changeTTL) {
             $ttl_set = int(rand(30) + 34);
             &ipt("-t mangle -I INetSim_$PID -s $SRC_IP -d $DST_IP -p $PROTO --dport $DST_PORT -m connmark --mark 0 -j CONNMARK --set-mark $ttl_set");
-            &INetSim::Log::SubLog("[$SRC_IP:$SRC_PORT] Forwarding $PROTO connections from host '$SRC_IP' ($MAC) to destination '$DST_IP:$DST_PORT', TTL set to $ttl_set.", $serviceName, $$);
+            &INetSim::Log::SubLog("[$SRC_IP:$SRC_PORT] Forwarding $PROTO connections from host '$SRC_IP'" . (($MAC) ? " ($MAC)" : "") . " to destination '$DST_IP:$DST_PORT', TTL set to $ttl_set.", $serviceName, $$);
         }
         else {
-            &INetSim::Log::SubLog("[$SRC_IP:$SRC_PORT] Forwarding $PROTO connections from host '$SRC_IP' ($MAC) to destination '$DST_IP:$DST_PORT'.", $serviceName, $$);
+            &INetSim::Log::SubLog("[$SRC_IP:$SRC_PORT] Forwarding $PROTO connections from host '$SRC_IP'" . (($MAC) ? " ($MAC)" : "") . " to destination '$DST_IP:$DST_PORT'.", $serviceName, $$);
         }
         return 1;
     }
@@ -656,10 +657,10 @@ sub process_packet_tcpudp {
             if ($changeTTL) {
                 $ttl_set = int(rand(30) + 34);
                 &ipt("-t mangle -I INetSim_$PID -s $SRC_IP -d $DST_IP -p $PROTO --dport $DST_PORT -m connmark --mark 0 -j CONNMARK --set-mark $ttl_set");
-                &INetSim::Log::SubLog("[$SRC_IP:$SRC_PORT] Redirecting $PROTO connections from host '$SRC_IP' ($MAC), destination changed from '$DST_IP:$DST_PORT' to '$my_ip:$real_port', TTL set to $ttl_set.", $serviceName, $$);
+                &INetSim::Log::SubLog("[$SRC_IP:$SRC_PORT] Redirecting $PROTO connections from host '$SRC_IP'" . (($MAC) ? " ($MAC)" : "") . ", destination changed from '$DST_IP:$DST_PORT' to '$my_ip:$real_port', TTL set to $ttl_set.", $serviceName, $$);
             }
             else {
-                &INetSim::Log::SubLog("[$SRC_IP:$SRC_PORT] Redirecting $PROTO connections from host '$SRC_IP' ($MAC), destination changed from '$DST_IP:$DST_PORT' to '$my_ip:$real_port'.", $serviceName, $$);
+                &INetSim::Log::SubLog("[$SRC_IP:$SRC_PORT] Redirecting $PROTO connections from host '$SRC_IP'" . (($MAC) ? " ($MAC)" : "") . ", destination changed from '$DST_IP:$DST_PORT' to '$my_ip:$real_port'.", $serviceName, $$);
             }
             return 1;
         }
@@ -674,10 +675,10 @@ sub process_packet_tcpudp {
                         if ($changeTTL) {
                             $ttl_set = int(rand(30) + 34);
                             &ipt("-t mangle -I INetSim_$PID -s $SRC_IP -d $DST_IP -p $PROTO --dport $DST_PORT -m connmark --mark 0 -j CONNMARK --set-mark $ttl_set");
-                            &INetSim::Log::SubLog("[$SRC_IP:$SRC_PORT] No redirect needed for $PROTO connections from host '$SRC_IP' ($MAC) to destination '$DST_IP:$DST_PORT', TTL set to $ttl_set.", $serviceName, $$);
+                            &INetSim::Log::SubLog("[$SRC_IP:$SRC_PORT] No redirect needed for $PROTO connections from host '$SRC_IP'" . (($MAC) ? " ($MAC)" : "") . " to destination '$DST_IP:$DST_PORT', TTL set to $ttl_set.", $serviceName, $$);
                         }
                         else {
-                            &INetSim::Log::SubLog("[$SRC_IP:$SRC_PORT] No redirect needed for $PROTO connections from host '$SRC_IP' ($MAC) to destination '$DST_IP:$DST_PORT'.", $serviceName, $$);
+                            &INetSim::Log::SubLog("[$SRC_IP:$SRC_PORT] No redirect needed for $PROTO connections from host '$SRC_IP'" . (($MAC) ? " ($MAC)" : "") . " to destination '$DST_IP:$DST_PORT'.", $serviceName, $$);
                         }
                         return 0;
                     }
@@ -690,10 +691,10 @@ sub process_packet_tcpudp {
                         if ($changeTTL) {
                             $ttl_set = int(rand(30) + 34);
                             &ipt("-t mangle -I INetSim_$PID -s $SRC_IP -d $DST_IP -p $PROTO --dport $DST_PORT -m connmark --mark 0 -j CONNMARK --set-mark $ttl_set");
-                            &INetSim::Log::SubLog("[$SRC_IP:$SRC_PORT] Redirecting $PROTO connections from host '$SRC_IP' ($MAC), destination changed from '$DST_IP:$DST_PORT' to '$my_ip:$DST_PORT', TTL set to $ttl_set.", $serviceName, $$);
+                            &INetSim::Log::SubLog("[$SRC_IP:$SRC_PORT] Redirecting $PROTO connections from host '$SRC_IP'" . (($MAC) ? " ($MAC)" : "") . ", destination changed from '$DST_IP:$DST_PORT' to '$my_ip:$DST_PORT', TTL set to $ttl_set.", $serviceName, $$);
                         }
                         else {
-                            &INetSim::Log::SubLog("[$SRC_IP:$SRC_PORT] Redirecting $PROTO connections from host '$SRC_IP' ($MAC), destination changed from '$DST_IP:$DST_PORT' to '$my_ip:$DST_PORT'.", $serviceName, $$);
+                            &INetSim::Log::SubLog("[$SRC_IP:$SRC_PORT] Redirecting $PROTO connections from host '$SRC_IP'" . (($MAC) ? " ($MAC)" : "") . ", destination changed from '$DST_IP:$DST_PORT' to '$my_ip:$DST_PORT'.", $serviceName, $$);
                         }
                         return 1;
                     }
@@ -708,10 +709,10 @@ sub process_packet_tcpudp {
                 if ($changeTTL) {
                     $ttl_set = int(rand(30) + 34);
                     &ipt("-t mangle -I INetSim_$PID -s $SRC_IP -d $DST_IP -p $PROTO --dport $DST_PORT -m connmark --mark 0 -j CONNMARK --set-mark $ttl_set");
-                    &INetSim::Log::SubLog("[$SRC_IP:$SRC_PORT] Redirecting $PROTO connections from host '$SRC_IP' ($MAC), destination changed from '$DST_IP:$DST_PORT' to '$my_ip:$dummyPortTCP', TTL set to $ttl_set.", $serviceName, $$);
+                    &INetSim::Log::SubLog("[$SRC_IP:$SRC_PORT] Redirecting $PROTO connections from host '$SRC_IP'" . (($MAC) ? " ($MAC)" : "") . ", destination changed from '$DST_IP:$DST_PORT' to '$my_ip:$dummyPortTCP', TTL set to $ttl_set.", $serviceName, $$);
                 }
                 else {
-                    &INetSim::Log::SubLog("[$SRC_IP:$SRC_PORT] Redirecting $PROTO connections from host '$SRC_IP' ($MAC), destination changed from '$DST_IP:$DST_PORT' to '$my_ip:$dummyPortTCP'.", $serviceName, $$);
+                    &INetSim::Log::SubLog("[$SRC_IP:$SRC_PORT] Redirecting $PROTO connections from host '$SRC_IP'" . (($MAC) ? " ($MAC)" : "") . ", destination changed from '$DST_IP:$DST_PORT' to '$my_ip:$dummyPortTCP'.", $serviceName, $$);
                 }
                 return 1;
             }
@@ -724,10 +725,10 @@ sub process_packet_tcpudp {
                 if ($changeTTL) {
                     $ttl_set = int(rand(30) + 34);
                     &ipt("-t mangle -I INetSim_$PID -s $SRC_IP -d $DST_IP -p $PROTO --dport $DST_PORT -m connmark --mark 0 -j CONNMARK --set-mark $ttl_set");
-                    &INetSim::Log::SubLog("[$SRC_IP:$SRC_PORT] Redirecting $PROTO connections from host '$SRC_IP' ($MAC), destination changed from '$DST_IP:$DST_PORT' to '$my_ip:$dummyPortUDP', TTL set to $ttl_set.", $serviceName, $$);
+                    &INetSim::Log::SubLog("[$SRC_IP:$SRC_PORT] Redirecting $PROTO connections from host '$SRC_IP'" . (($MAC) ? " ($MAC)" : "") . ", destination changed from '$DST_IP:$DST_PORT' to '$my_ip:$dummyPortUDP', TTL set to $ttl_set.", $serviceName, $$);
                 }
                 else {
-                    &INetSim::Log::SubLog("[$SRC_IP:$SRC_PORT] Redirecting $PROTO connections from host '$SRC_IP' ($MAC), destination changed from '$DST_IP:$DST_PORT' to '$my_ip:$dummyPortUDP'.", $serviceName, $$);
+                    &INetSim::Log::SubLog("[$SRC_IP:$SRC_PORT] Redirecting $PROTO connections from host '$SRC_IP'" . (($MAC) ? " ($MAC)" : "") . ", destination changed from '$DST_IP:$DST_PORT' to '$my_ip:$dummyPortUDP'.", $serviceName, $$);
                 }
                 return 1;
             }
@@ -735,90 +736,13 @@ sub process_packet_tcpudp {
     else {
         # mark rule for more packets with the same properties
         &ipt("-t mangle -I INetSim_$PID -s $SRC_IP -d $DST_IP -p $PROTO --dport $DST_PORT -j MARK --set-mark 1");
-        &INetSim::Log::SubLog("[$SRC_IP:$SRC_PORT] No rule for $PROTO connections from host '$SRC_IP' ($MAC) to destination '$DST_IP:$DST_PORT' - ignored.", $serviceName, $$);
+        &INetSim::Log::SubLog("[$SRC_IP:$SRC_PORT] No rule for $PROTO connections from host '$SRC_IP'" . (($MAC) ? " ($MAC)" : "") . " to destination '$DST_IP:$DST_PORT' - ignored.", $serviceName, $$);
     }
     return 0;
 }
 
 
 
-sub close_queue {
-    $ipq->close();
-}
-
-
-
-sub process_queue {
-    while () {
-        my $msg = $ipq->get_message();
-        if (!defined $msg) {
-            next if IPTables::IPv4::IPQueue->errstr eq 'Timeout';
-        }
-        my $mac = $msg->hw_addr();
-        my $ip_packet = $msg->payload();
-        $IN_DEV = $msg->indev_name();
-        $OUT_DEV = $msg->outdev_name();
-        my $mark = $msg->mark();
-        my $new_packet;
-        my $changed = 0;
-
-        if (! $mark) {
-            &split_mac($mac);
-            &split_ip($ip_packet);
-            if ($PROTO eq "tcp") {
-                &split_tcp($IP{data});
-                &process_packet_tcpudp;
-            }
-            elsif ($PROTO eq "udp") {
-                &split_udp($IP{data});
-                &process_packet_tcpudp;
-            }
-            elsif ($PROTO eq "icmp") {
-                &split_icmp($IP{data});
-                &process_packet_icmp;
-                # handle icmp timestamp replies, if configured
-                if ($icmp_ts && $TYPE == 14) {
-                    my $ts_reply = &fake_ts_reply;
-                    if ($ts_reply) {
-                        $ICMP{receive} = $ts_reply;
-                        $ICMP{transmit} = $ts_reply;
-                        $IP{data} = &build_icmp;
-                        $new_packet = &build_ip;
-                        $changed = 1;
-                    }
-                }
-            }
-        }
-        elsif ($icmp_ts && $mark == 2) {
-            &split_mac($mac);
-            &split_ip($ip_packet);
-            if ($PROTO eq "icmp") {
-                &split_icmp($IP{data});
-                if ($TYPE == 14) {
-                    my $ts_reply = &fake_ts_reply;
-                    if ($ts_reply) {
-                        $ICMP{receive} = $ts_reply;
-                        $ICMP{transmit} = $ts_reply;
-                        $IP{data} = &build_icmp;
-                        $new_packet = &build_ip;
-                        $changed = 1;
-                    }
-                }
-            }
-        }
-
-        if (! $changed) {
-            $ipq->set_verdict($msg->packet_id, NF_ACCEPT) or die IPTables::IPv4::IPQueue->errstr;
-        }
-        else {
-            $ipq->set_verdict($msg->packet_id, NF_ACCEPT, length ($new_packet), $new_packet) or die IPTables::IPv4::IPQueue->errstr;
-        }
-    }
-    $ipq->close();
-}
-
-
-
 sub fake_ts_reply {
     ($icmp_ts) or return;
     if ($icmp_ts == 1) {
@@ -1218,26 +1142,109 @@ sub build_tcp {
 
 
 
+sub close_queue {
+    $nfq->unbind(AF_INET);
+    $nfq->close();
+}
+
+
+
+sub process_queue {
+    my $payload = shift;
+
+    if ($payload) {
+	#my $mac = $payload->get_packet_hw();
+	my $ip_packet = $payload->get_data();
+	$IN_DEV = $payload->get_indev();
+	$OUT_DEV = $payload->get_outdev();
+	my $mark = $payload->get_nfmark();
+	my $new_packet;
+	my $changed = 0;
+
+        if (! $mark) {
+            #split_mac($mac);
+            split_ip($ip_packet);
+            if ($PROTO eq "tcp") {
+                split_tcp($IP{data});
+                process_packet_tcpudp();
+            }
+            elsif ($PROTO eq "udp") {
+                split_udp($IP{data});
+                process_packet_tcpudp();
+            }
+            elsif ($PROTO eq "icmp") {
+                split_icmp($IP{data});
+                process_packet_icmp();
+                # handle icmp timestamp replies, if configured
+                if ($icmp_ts && $TYPE == 14) {
+                    my $ts_reply = fake_ts_reply();
+                    if ($ts_reply) {
+                        $ICMP{receive} = $ts_reply;
+                        $ICMP{transmit} = $ts_reply;
+                        $IP{data} = build_icmp();
+                        $new_packet = build_ip();
+                        $changed = 1;
+                    }
+                }
+            }
+        }
+        elsif ($icmp_ts && $mark == 2) {
+            #split_mac($mac);
+            split_ip($ip_packet);
+            if ($PROTO eq "icmp") {
+                split_icmp($IP{data});
+                if ($TYPE == 14) {
+                    my $ts_reply = fake_ts_reply();
+                    if ($ts_reply) {
+                        $ICMP{receive} = $ts_reply;
+                        $ICMP{transmit} = $ts_reply;
+                        $IP{data} = build_icmp();
+                        $new_packet = build_ip();
+                        $changed = 1;
+                    }
+                }
+            }
+        }
+
+        if (! $changed) {
+            die ("set verdict failed: $!") if ($payload->set_verdict($nfqueue::NF_ACCEPT) < 0);
+        }
+        else {
+            die("set verdict failed for modified packet") if ($payload->set_verdict_modified($nfqueue::NF_ACCEPT, $new_packet, length($new_packet)) < 0);
+        }
+    }
+
+    return;
+}
+
+
+
 sub run {
     $0 = "inetsim [$serviceName]";
     $SIG{'INT'} = $SIG{'HUP'} = $SIG{'PIPE'} = $SIG{'QUIT'} = $SIG{'TERM'} = 'IGNORE';
     $ENV{PATH} = "/usr/local/sbin:/usr/local/bin:/usr/sbin:/usr/bin:/sbin:/bin";
     # check for uid=0 and iptables
-    &check_requirements;
-    # check - is ipqueue runnable ?
+    check_requirements();
+    # check - is nfqueue runnable ?
     eval {
-           $ipq = new IPTables::IPv4::IPQueue(copy_mode => IPQ_COPY_PACKET, copy_range => 1500) or die IPTables::IPv4::IPQueue->errstr;
+	$nfq = new nfqueue::queue() or die "setup nfqueue failed: $!\n";
     };
     # isn't => exit
     if ($@) {
-        &INetSim::Log::MainLog("failed! Error: $@", $serviceName);
+        INetSim::Log::MainLog("failed! Error: $@", $serviceName);
         exit 1;
     }
-    &parse_static_rules;
-    &INetSim::Log::MainLog("started (PID $$)", $serviceName);
-    $SIG{'TERM'} = sub { &delete_chains; &close_queue; &INetSim::Log::MainLog("stopped (PID $$)", $serviceName); exit 0;};
-    &create_chains;
-    &process_queue;
+    parse_static_rules();
+    INetSim::Log::MainLog("started (PID $$)", $serviceName);
+    create_chains();
+    # set callback
+    $nfq->set_callback(\&process_queue);
+    # open nfqueue
+    my $res = $nfq->fast_open(0, AF_INET);
+    die("open nfqueue failed: $!\n") if ($res < 0);
+    $SIG{'TERM'} = sub { delete_chains(); close_queue(); INetSim::Log::MainLog("stopped (PID $$)", $serviceName); exit 0;};
+    # run nfqueue loop
+    $nfq->try_run();
 }
 
 
@@ -1247,6 +1254,9 @@ sub run {
 #
 # History:
 #
+# Version 0.21  (2017-11-23) me
+# - modified to work with newer kernel versions
+#
 # Version 0.20  (2010-04-10) me
 # - bugfix: change all timestamp replies, not just the first packet
 # - some small changes in functions check_requirements() and ipt()
